from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np

from ...common import Model, TX, TY
from ...utils import TX_SCALE, TY_SCALE


class Model2D(Model):
    r"""Combines a pair of :class:`~ionics_fits.common.Model`\'s, each of which is a
    function of a single x-dimension, to make a new :class:`~ionics_fits.common.Model`\,
    which is a function of 2 x-axis dimensions.

    All y-axis data is generated from the output of the first model; the output from
    the second model provides the values of certain "result" parameters used by the
    first model. In other words::

      model_0 = models[0] = f(x_axis_0)
      model_1 = models[1] = g(x_axis_1)
      y(x_0, x_1) = model_0(x_0 | result_params = model_1(x_1))

    The 2D models generated using this class are separable into functions of the two
    x-axes. As a result, it can only generate axis-aligned models.

    This model provides a quick and convenient means of extending the x-axis
    dimensionality of existing models. The design aim is to maximise flexibility while
    making minimal assumptions about the underlying
    :class:`~ionics_fits.common.Model`\s.

    Model parameters and results:

    * All parameters from the two models - apart from the first model's *result
      parameters* - are parameters of the 2D model.
    * All derived results from the two models are included in the :class:`Model2D`\'s
      derived results.
    * A parameter/derived result named ``param`` from a model named ``model`` is
      exposed as a parameter / result of the :class:`Model2D` named ``param_model``.
    * A :class:`~ionics_fits.models.transformations.mapped_model.MappedModel` can be
      used to provide custom naming schemes

    Example usage:

    .. doctest::

        from ionics_fits.models.gaussian import Gaussian
        from ionics_Fits.models.transformations.model_2d import Model2D

        Gaussian2D = Model2D(
            models=(Gaussian(), Gaussian),
            result_params=("a_x0",),
        )

    This fits a Gaussian model along the first x-axis dimension. It then takes the
    amplitudes of those Gaussians at each value of the second x-axis dimension and
    fits those to a Gaussian to produce a 2D Gaussian fit.

    At present this class aggregates a pair of 1D models to make a 2D model. It would
    be easy to extend it to the more general case of aggregating an arbitrary number of
    models, each with arbitrary x-axis dimensionality. This has not been done yet
    because there has not been a use-case.

    See also :py:mod:`~ionics_fits.models.multi_x`.
    """

    def __init__(
        self,
        models: Tuple[Model, Model],
        result_params: Tuple[str],
        model_names: Optional[Tuple[str, str]] = None,
    ):
        r"""
        :param models: Tuple containing the two :class:`~ionics_fits.common.Model`\s to
            be combined to make the 2D model. The model instances are considered "owned"
            by the 2D model (they are not copied). They should not be referenced
            externally.
        :param result_params: tuple of names of "result parameters" of the first model,
            whose value is found by evaluating the second model. The order of parameter
            names in this tuple must match the order of y-axis dimensions for the second
            model.
        :param model_names: optional tuple of names for the two models. These are used
            when generating names for fit results and derived parameters. Empty strings
            are allowed so long as the two models do not share any parameter names. If
            this argument is ``None`` the model names default to ``x0`` and ``x1``
            respectively. If a model name is an empty string, the trailing underscore
            is omitted in parameter / result names.
        """
        self.models = models
        self.result_params = result_params
        self.model_names = model_names if model_names is not None else ("x0", "x1")

        missing = set(self.result_params) - set(self.models[0].parameters.keys())
        if missing:
            raise ValueError(
                "Result parameters must be parameters of the first model. Unrecognised "
                f"result parameter names are: {missing}"
            )

        if len(self.result_params) != self.models[1].get_num_y_axes():
            raise ValueError(
                f"{len(self.result_params)} result parameters passed when second model "
                f"requires {self.models[1].get_num_y_axes()}"
            )

        if any([model.get_num_x_axes() != 1 for model in self.models]):
            raise ValueError("Model2D currently only supports 1D models as inputs")

        # Map the parameters of the 1D models onto parameters of the 2D models
        parameters = {}
        internal_parameters = [
            self.models[0].parameters[param_name] for param_name in self.result_params
        ]

        self.model_param_maps = ({}, {})  # model_param_name: new_param_name

        for model_idx, model_name in enumerate(self.model_names):
            # Result parameters are internal parameters of the 2D model
            exclusions = self.result_params if model_idx == 0 else []
            suffix = f"_{model_name}" if model_name else ""
            self.model_param_maps[model_idx].update(
                {
                    model_param_name: f"{model_param_name}{suffix}"
                    for model_param_name in self.models[model_idx].parameters.keys()
                    if model_param_name not in exclusions
                }
            )

        duplicates = set(self.model_param_maps[0].values()).intersection(
            set(self.model_param_maps[1].values())
        )
        if duplicates:
            raise ValueError(
                f"Duplicate parameter names found between the two models: {duplicates}."
            )

        # Create the parameters / internal parameters for the 2D model
        for model_idx, param_map in enumerate(self.model_param_maps):
            model = self.models[model_idx]
            parameters.update(
                {
                    new_param_name: model.parameters[model_param_name]
                    for model_param_name, new_param_name in param_map.items()
                }
            )
            internal_parameters += model.internal_parameters

        super().__init__(parameters=parameters, internal_parameters=internal_parameters)
        self.wrap_scale_funcs()

    def wrap_scale_funcs(self):
        r"""Called during
        :meth:`~ionics_fits.models.transformations.model_2d.Model2D.__init__` to make
        sure that each :class:`~ionics_fits.common.ModelParameter` uses the scale factor
        from the appropriate x-axis when rescaling.

        This method wraps the :attr:`~ionics_fits.common.ModelParameter.scale_func`\s
        for each parameter of the 1D models to use the x-axis scale factor for their
        x-axis dimension when rescaling.

        This is appropriate where each model parameter scales with just that the x-axis
        associated with its model. If this is not the case, this method should be
        overridden. See :class:`~ionics_fits.models.multi_x.Gaussian2D` for an example
        of this.
        """

        def wrapped_scale_func(
            scale_func: Callable[[TX_SCALE, TY_SCALE], float],
            x_ax: int,
            x_scales: TX_SCALE,
            y_scales: TY_SCALE,
        ) -> float:
            return scale_func(np.array([x_scales[x_ax]]), y_scales)

        for model_idx, model in enumerate(self.models):
            parameters = list(model.parameters.values()) + model.internal_parameters
            for parameter in parameters:
                scale_func = parameter.scale_func
                parameter.scale_func = partial(
                    wrapped_scale_func, scale_func, model_idx
                )
                parameter.scale_func.__name__ = scale_func.__name__

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        rescale_x0, rescale_y0 = self.models[0].can_rescale()
        rescale_x1, rescale_y1 = self.models[1].can_rescale()

        rescale_x = list(np.hstack((rescale_x0, rescale_x1)))
        rescale_y = np.all(np.vstack((rescale_y0, rescale_y1)), axis=0)

        return rescale_x, rescale_y

    def get_num_x_axes(self) -> int:
        return 2

    def get_num_y_axes(self) -> int:
        return self.models[0].get_num_y_axes()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        x = np.atleast_2d(x)

        model_0_values = {
            model_param_name: param_values[new_param_name]
            for model_param_name, new_param_name in self.model_param_maps[0].items()
        }
        model_1_values = {
            model_param_name: param_values[new_param_name]
            for model_param_name, new_param_name in self.model_param_maps[1].items()
        }

        x_1_ax = np.unique(x[1, :])
        result_param_values = self.models[1].func(x_1_ax, model_1_values)
        result_param_values = np.atleast_2d(result_param_values)

        y = np.zeros((self.get_num_y_axes(), x.shape[1]), dtype=np.float64)
        for x_1_idx, x_1 in np.ndenumerate(x_1_ax):
            x_0_mask = x[1, :] == x_1
            x_0_ax = x[0, x_0_mask]
            model_0_values.update(
                {
                    param_name: float(result_param_values[param_idx, x_1_idx])
                    for param_idx, param_name in enumerate(self.result_params)
                }
            )
            y[:, x_0_mask] = np.atleast_2d(self.models[0].func(x_0_ax, model_0_values))

        return y

    def estimate_parameters(self, x: TX, y: TY):
        x_1_ax = np.unique(x[1, :])

        # Step 1: find heuristics for model 0 for each point along the second x-axis
        x_1_keep = np.full(len(x_1_ax) + 1, True, dtype=bool)
        heuristics = {
            param_name: np.zeros(len(x_1_ax) + 1, dtype=np.float64)
            for param_name in self.models[0].parameters.keys()
        }

        for x_1_idx, x_1 in np.ndenumerate(x_1_ax):
            x_1_mask = x[1, :] == x_1
            x_0_ax = x[0, x_1_mask]
            y_x_1 = y[:, x_1_mask]

            self.models[0].clear_heuristics()
            try:
                self.models[0].estimate_parameters(x=x_0_ax, y=y_x_1)
            except Exception:
                x_1_keep[x_1_idx] = False

            for param_name, param_data in self.models[0].parameters.items():
                heuristics[param_name][x_1_idx] = param_data.get_initial_value()

        x_1_ax = x_1_ax[x_1_keep[:-1]]
        for param_name in heuristics.keys():
            heuristics[param_name] = heuristics[param_name][x_1_keep]

        for param_name in heuristics.keys():
            heuristics[param_name][-1] = np.mean(heuristics[param_name][:-1])

        # Step 2: pick the best heuristic
        costs = np.zeros(len(x_1_ax) + 1, dtype=np.float64)
        for heuristic_idx in range(len(costs)):
            y_heuristic = np.zeros_like(y)
            x_1_heuristics = {
                param_name: param_heuristic[heuristic_idx]
                for param_name, param_heuristic in heuristics.items()
            }

            for x_1_idx, x_1 in np.ndenumerate(x_1_ax):
                x_1_heuristics.update(
                    {
                        param_name: heuristics[param_name][x_1_idx]
                        for param_name in self.result_params
                    }
                )

                x_1_mask = x[1, :] == x_1
                x_0_ax = x[0, x_1_mask]
                y_heuristic[:, x_1_mask] = self.models[0].func(
                    x=x_0_ax, param_values=x_1_heuristics
                )
            costs[heuristic_idx] = np.sqrt(np.sum((y - y_heuristic) ** 2))

        lowest_cost = np.argmin(costs)
        best_heuristic = {
            param_name: heuristics[param_name][lowest_cost]
            for param_name in heuristics.keys()
        }

        for model_param_name, new_param_name in self.model_param_maps[0].items():
            param_heuristic = best_heuristic[model_param_name]
            self.parameters[new_param_name].heuristic = param_heuristic

        # Step 3: use the result parameter heuristics from model 0 as an input to
        # model 1
        y_1 = np.zeros((len(self.result_params), len(x_1_ax)), dtype=np.float64)
        for param_idx, result_param_name in enumerate(self.result_params):
            y_1[param_idx, :] = heuristics[result_param_name][:-1]

        self.models[1].estimate_parameters(x_1_ax, y_1)

        for model_param_name, new_param_name in self.model_param_maps[1].items():
            param_data = self.models[1].parameters[model_param_name]
            heuristic = param_data.get_initial_value()
            self.parameters[new_param_name].heuristic = heuristic

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_values = {}
        derived_uncertainties = {}

        model_values = ({}, {})
        model_uncertainties = ({}, {})
        model_derived_values = ({}, {})
        model_derived_uncertainties = ({}, {})

        # Result parameters do not have meaningful fitted values / uncertainties
        for param_name in self.result_params:
            model_values[0][param_name] = float("nan")
            model_uncertainties[0][param_name] = float("nan")

        for model_idx, model in enumerate(self.models):
            param_map = self.model_param_maps[model_idx]
            model_values[model_idx].update(
                {
                    model_param_name: fitted_params[new_param_name]
                    for model_param_name, new_param_name in param_map.items()
                }
            )
            model_uncertainties[model_idx].update(
                {
                    model_param_name: fit_uncertainties[new_param_name]
                    for model_param_name, new_param_name in param_map.items()
                }
            )

            derived = model.calculate_derived_params(
                x=x[model_idx, :],
                y=y,
                fitted_params=model_values[model_idx],
                fit_uncertainties=model_uncertainties[model_idx],
            )

            if self.model_names[model_idx] != "":
                model_suffix = f"_{self.model_names[model_idx]}"
            else:
                model_suffix = ""

            model_derived_values[model_idx].update(
                {
                    f"{param_name}{model_suffix}": value
                    for param_name, value in derived[0].items()
                }
            )
            model_derived_uncertainties[model_idx].update(
                {
                    f"{param_name}{model_suffix}": value
                    for param_name, value in derived[1].items()
                }
            )

        duplicates = set(model_derived_values[0].keys()).intersection(
            set(model_derived_values[1].keys())
        )
        if duplicates:
            raise ValueError(
                "Duplicate derived result names found between the two models: "
                f"{duplicates}."
            )

        derived_values = {}
        derived_values.update(model_derived_values[0])
        derived_values.update(model_derived_values[1])

        derived_uncertainties = {}
        derived_uncertainties.update(model_derived_uncertainties[0])
        derived_uncertainties.update(model_derived_uncertainties[1])

        return derived_values, derived_uncertainties



import numpy as np
from scipy import special
from typing import TYPE_CHECKING

from ..common import Array, ModelParameter
from ..utils import scale_invariant

if TYPE_CHECKING:
    num_fock_states = float


# pytype: disable=invalid-annotation
def thermal_state_probs(
    n_max: int, n_bar: ModelParameter(lower_bound=0, scale_func=scale_invariant)
) -> Array[("num_fock_states",), np.float64]:
    """Thermal state probability distribution.

    :param n_max: the distribution is truncated at a maximum Fock state of ``|n_max>``
    :param n_bar: the mean Fock state occupation
    :returns: array of Fock state occupation probabilities
    """
    n = np.arange(n_max + 1, dtype=int)
    return ((n_bar / (n_bar + 1)) ** n) / (n_bar + 1)


# pytype: enable=invalid-annotation

# pytype: disable=invalid-annotation
def coherent_state_probs(
    n_max: int, alpha: ModelParameter(lower_bound=0, scale_func=scale_invariant)
) -> Array[("num_fock_states",), np.float64]:
    """Coherent state probability distribution.

    A coherent state is defined as::

        |α> = exp(α a_dag - α* a) |0>

    where ``a_dag`` and ``a`` denote the harmonic-oscillator creation and annihilation
    operators. The mean Fock state occupancy is given by ``n_bar = |α|^2``

    :param n_max: the distribution is truncated at a maximum Fock state of ``|n_max>``
    :param alpha: Complex displacement parameter
    :returns: array of Fock state occupation probabilities
    """
    n = np.arange(n_max + 1, dtype=int)
    n_bar = np.abs(alpha) ** 2

    # The standard expression is: P_n = n_bar**n * exp(-n_bar) / n!
    # However, this is difficult to evaluate as n increases, so we use an alternate form
    # NB for integer n: gamma(n + 1) = n!
    if n_bar == 0:
        P_n = np.zeros_like(n)
        P_n[0] = 1
    else:
        P_n = np.exp(n * np.log(n_bar) - n_bar - special.gammaln(n + 1))
    return P_n


# pytype: enable=invalid-annotation


# pytype: disable=invalid-annotation
def squeezed_state_probs(
    n_max: int, zeta: ModelParameter(lower_bound=0, scale_func=scale_invariant)
) -> Array[("num_fock_states",), np.float64]:
    """Squeezed state probability distribution.

    A pure squeezed state is defined as::

        |ζ> = exp[1 / 2 *  (ζ* a^2 - ζ a_dag^2)] |0>,

    where ``a_dag`` and ``a`` denote the harmonic-oscillator creation and annihilation
    operators.

    :param n_max: the distribution is truncated at a maximum Fock state of ``|n_max>``
    :param zeta: Complex squeezing parameter
    :returns: array of Fock state occupation probabilities
    """
    n = np.arange(n_max + 1, dtype=int)

    r = np.abs(zeta)
    if r == 0:
        P_n = np.zeros_like(n)
        P_n[0] = 1
    else:
        P_n = (
            (np.tanh(r) ** (2 * n))
            / np.cosh(r)
            * np.exp(
                special.gammaln(2 * n + 1)
                - 2 * n * np.log(2)
                - 2 * special.gammaln(n + 1)
            )
        )
        # For a squeezed state, only even Fock states are occupied
        P_n[1::2] = 0

    return P_n


# pytype: enable=invalid-annotation



from typing import Dict, List, Tuple

import numpy as np

from . import heuristics
from .transformations.mapped_model import MappedModel
from ..common import Model, ModelParameter, TX, TY
from ..utils import (
    scale_invariant,
    scale_power,
    scale_undefined,
    scale_x,
    scale_y,
    TX_SCALE,
    TY_SCALE,
)


class Power(Model):
    """Single-power fit according to::

        y = a*(x-x0)^n + y0

    ``x - x0`` must always be strictly greater than ``0``. This is because ``n`` can
    take non-integral values (for integer coefficients use
    :class:`~ionics_fits.models.polynomial.Polynomial` instead) and this function's
    return is real-valued.

    The fit will often struggle when both ``y0`` and ``n`` are floated if the dataset
    doesn't contain some asymptotic values where ``y ~ y0``. The more you can help it
    out by bounding parameters and providing initial guesses the better.

    The fit will generally struggle to converge if both ``a`` and ``y0`` are floated
    unless it is given some guidance (e.g. initial values).
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        rescale_x = False if self.parameters["n"].fixed_to is None else True
        return [rescale_x], [True]

    def rescale(self, x_scales: TX_SCALE, y_scales: TY_SCALE):
        def scale_func(x_scales, y_scales) -> float:
            # NB the common scale functions do not support float powers
            x_scale = float(x_scales)
            y_scale = float(y_scales)
            if x_scales == 1.0:
                return y_scale
            return y_scale / np.float_power(x_scale, self.parameters["n"].fixed_to)

        self.parameters["a"].scale_func = scale_func
        super().rescale(x_scales=x_scales, y_scales=y_scales)

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        a: ModelParameter(fixed_to=1, scale_func=scale_undefined),
        x0: ModelParameter(fixed_to=0, scale_func=scale_x()),
        y0: ModelParameter(scale_func=scale_y()),
        n: ModelParameter(scale_func=scale_invariant),
    ) -> TY:
        """
        * ``a``: y-axis scale factor
        * ``x0``: x-axis offset
        * ``y0``: y-axis offset
        * ``n``: power
        """
        assert np.all(x - x0 >= 0), "`x - x0` must be > 0"
        return a * np.float_power(x - x0, n) + y0

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        unknowns = {
            param
            for param, param_data in self.parameters.items()
            if param_data.fixed_to is None
        }

        self.parameters["x0"].heuristic = 0.0
        self.parameters["y0"].heuristic = 0.0
        self.parameters["a"].heuristic = 1.0
        self.parameters["n"].heuristic = 1.0

        x0 = self.parameters["x0"].get_initial_value()
        y0 = self.parameters["y0"].get_initial_value()
        a = self.parameters["a"].get_initial_value()
        n = self.parameters["n"].get_initial_value()

        if len(unknowns) == 0 or "x0" in unknowns or len(unknowns) > 2:
            # No heuristic needed / we don't have a good heuristic for this case
            # Fit may well fail if we hit this :(
            pass

        elif unknowns == {"a"}:
            raise ValueError
            x = x - x0
            y = y - y0
            a = y / np.float_power(x, n)
            a = heuristics.param_min_sqrs(
                model=self, x=x, y=y, scanned_param="a", scanned_param_values=a
            )[0]
            self.parameters["a"].heuristic = a

        elif unknowns == {"n"}:
            n = self.optimal_n(x, y)[0]
            self.parameters["n"].heuristic = n

        elif unknowns == {"y0"}:
            y0 = y - a * np.float_power(x - x0, n)
            y0 = heuristics.param_min_sqrs(
                model=self, x=x, y=y, scanned_param="y0", scanned_param_values=y0
            )[0]
            self.parameters["y0"].heuristic = y0

        elif "a" in unknowns:
            pass  # don't have a great heuristic for these cases

        else:
            assert unknowns == {"y0", "n"}, unknowns

            # Datasets normally taken such that they contain a value of y close to y0
            y0_guesses = np.array(
                [
                    0,
                    np.min(y),
                    np.max(y),
                    np.mean(y),
                ]
            )
            y0_bounds = [
                self.parameters["y0"].lower_bound,
                self.parameters["y0"].upper_bound,
            ]
            y0_bounds = [bound for bound in y0_bounds if np.isfinite(bound)]
            y0_guesses = np.append(y0_guesses, y0_bounds)

            ns = np.zeros_like(y0_guesses)
            costs = np.zeros_like(y0_guesses)
            for idx, y0 in np.ndenumerate(y0_guesses):
                self.parameters["y0"].heuristic = y0
                ns[idx], costs[idx] = self.optimal_n(x, y)

            self.parameters["n"].heuristic = float(ns[np.argmin(costs)])
            self.parameters["y0"].heuristic = float(y0_guesses[np.argmin(costs)])

    def optimal_n(self, x, y):
        """Find the optimal (in the least-squared residuals sense) value of `n`
        based on our current best guesses for the other parameters.

        We use the fact that if `y = x^n` then `n = log(y) / log(x)`. This gives
        us an estimate for `n` at each value of x. We choose the one which results
        in lowest sum of squares residuals.
        """
        if self.parameters["n"].has_user_initial_value():
            return self.parameters["n"].get_initial_value(), 0

        x0 = self.parameters["x0"].get_initial_value()
        y0 = self.parameters["y0"].get_initial_value()
        a = self.parameters["a"].get_initial_value()

        x_pr = x - x0
        y_pr = y - y0
        y_pr = y_pr / a

        # Avoid divide by zero errors
        valid = np.argwhere(np.logical_and(y_pr != 0, x_pr != 1))

        if len(valid) == 0:
            return 0, np.inf

        n = np.log(np.abs(y_pr[valid])) / np.log(np.abs(x_pr[valid]))
        n = n.squeeze()

        # Don't look for silly values of n
        n_min = max(-10, self.parameters["n"].lower_bound)
        n_max = min(10, self.parameters["n"].upper_bound)

        n = n[np.argwhere(np.logical_and(n >= n_min, n <= n_max))]

        if len(n) == 0:
            return 1, np.inf

        return heuristics.param_min_sqrs(self, x, y, "n", n)


def poly_fit_parameter(n):
    return ModelParameter(
        fixed_to=None,
        scale_func=scale_power(x_power=-n, y_power=1),
    )


def _generate_poly_parameters(poly_degree):
    params = {f"a_{n}": poly_fit_parameter(n) for n in range(poly_degree + 1)}
    params.update({"x0": ModelParameter(fixed_to=0, scale_func=scale_x())})
    return params


class Polynomial(Model):
    """Polynomial fit model according to::

        y = sum(a_n*(x-x0)^n) for n ={0...poly_degree}

    Model parameters:

    * ``a_0`` ... ``a_{poly_degree}``: polynomial coefficients
    * ``x0``: x-axis offset (fixed to 0 by default). Floating ``x0`` as well as
      polynomial coefficients results in an under-defined problem.
    """

    def __init__(self, poly_degree=10):
        """
        :param poly_degree: The degree of the polynomial that we're fitting
           defaults to 10.
        """
        self.poly_degree = poly_degree
        super().__init__(parameters=_generate_poly_parameters(poly_degree))

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        rescale_x = True if self.parameters["x0"].fixed_to == 0.0 else False
        return [rescale_x], [True]

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        x0 = param_values["x0"]
        p = np.array(
            [param_values[f"a_{n}"] for n in range(self.poly_degree, -1, -1)],
            dtype=np.float64,
        )

        y = np.polyval(p, x - x0)
        return y

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        self.parameters["x0"].heuristic = 0.0
        x0 = self.parameters["x0"].get_initial_value()

        free = [
            n
            for n in range(self.poly_degree + 1)
            if self.parameters[f"a_{n}"].fixed_to != 0.0
        ]
        if len(free) == 0:
            return

        deg = max(free)

        p = np.polyfit(x - x0, y, deg)
        for idx in range(deg + 1):
            self.parameters[f"a_{idx}"].heuristic = p[deg - idx]


class Line(MappedModel):
    """Straight line fit according to::

        y = a * x + y0

    Fit parameters (all floated by default unless stated otherwise):

    * ``y0``: y-axis intercept
    * ``a``: slope
    """

    def __init__(self):
        super().__init__(
            Polynomial(1),
            {"a": "a_1", "y0": "a_0"},
            {"x0": 0},
        )


class Parabola(MappedModel):
    """Parabola fit according to:
    `y = k * (x - x0)^2 + y0`

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis intercept
      - k: curvature

    Derived parameters:
        None
    """

    def __init__(self):
        inner = Polynomial(2)
        inner.parameters["x0"].fixed_to = None
        inner.parameters["a_2"].fixed_to = None
        super().__init__(
            inner,
            {"k": "a_2", "y0": "a_0", "x0": "x0"},
            {"a_1": 0},
        )

    def estimate_parameters(self, x: TX, y: TY):
        """
        If ``x0`` is floated, we map the Polynomial ``a_1`` coefficient onto a value for
        ``x0`` according to::

            y = a_0 + a_2 * x^2
            x -> x - x0: y = a_0 + a_2 * (x - x0)^2
            y = a_0 + a_2*x^2 + a_2 * x0^2 + 2*a_2*x*x0
            y = (a_0 + a_2 * x0^2) + 2*a_2*x*x0 + a_2*x^2

            a_0 -> a_0 + a_2 * x0^2
            a_1 -> 2*a_2*x0 => x0 = a_1/(2*a_2)
            a_2 -> a_2

        """
        super().estimate_parameters(x, y)

        a_0 = self.model.parameters["a_0"].get_initial_value()
        a_1 = self.model.parameters["a_1"].heuristic
        a_2 = self.model.parameters["a_2"].get_initial_value()

        x0 = -a_1 / (2 * a_2)
        self.parameters["x0"].heuristic = x0
        self.parameters["y0"].heuristic = a_0 - a_2 * x0**2



from typing import List, Tuple
import numpy as np

from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_x, scale_x_inv, scale_y


class LogisticFunction(Model):
    """Logistic function model according to::

        y = a / (1 + exp(-k*(x - x0))) + y0

    See :meth:`_func` for parameters.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        a: ModelParameter(scale_func=scale_y()),
        y0: ModelParameter(scale_func=scale_y()),
        x0: ModelParameter(scale_func=scale_x()),
        k: ModelParameter(scale_func=scale_x_inv(), lower_bound=0),
    ) -> TY:
        """
        :param a: y-axis scale factor
        :param y0: y-axis offset
        :param x0: x-axis offset
        :param k: logistic growth rate (steepness of the curve)
        """
        return a / (1 + np.exp(-k * (x - x0))) + y0

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        sgn = +1 if y[-1] > y[0] else -1
        y0 = self.parameters["y0"].get_initial_value(y[0])
        self.parameters["y0"].heuristic = y0

        a = self.parameters["a"].get_initial_value(sgn * y.ptp())
        self.parameters["a"].heuristic = a

        y_centre = y0 + a / 2
        x0 = self.parameters["x0"].get_initial_value(x[np.argmin(np.abs(y - y_centre))])
        self.parameters["x0"].heuristic = x0

        FWHMH_pts = np.abs(y - y_centre) < np.abs(a / 4)
        x_FWHMH = x[FWHMH_pts]
        FWHMH = x_FWHMH.ptp()

        k = self.parameters["k"].get_initial_value(1.098 / (FWHMH / 2))
        self.parameters["k"].heuristic = k



from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np

from ..utils import param_like
from ...common import Model, TX, TY
from ...utils import TX_SCALE, TY_SCALE


class RepeatedModel(Model):
    """Model formed by repeating a :class:`~ionics_fits.common.Model` one or more models
    along the y-axis to produce a new model, with greater y-axis dimensionality.

    The number of y-axis dimensions in the ``RepeatedModel`` is the product of the
    number of y-axis dimensions form the original model and the number of repetitions.

    Repeated models allow multiple datasets to be analysed simultaneously. This is
    useful, for example, when doing joint fits to datasets (using common parameters).

    The ``RepeatedModel`` has the following parameters and fit results:

    * all common parameters of the repeated model are parameters of the new model
    * for each independent (not common) parameter of the repeated model, ``param``,
      the ``RepeatedModel`` has parameters ``param_{n}`` for ``n`` in
      ``[0, .., num_repitions-1]``
    * for each independent parameter, ``param``, the ``RepeatedModel`` model
      has additional fit results representing statistics between the repetitions.
      For each independent parameter ``param``, the results dictionary will have
      additional quantities ``param_mean`` and ``param_peak_peak``.

    If ``aggregate_results`` is enabled the rules for fit results are modified as
    follows:

    * no statistical quantities are calculated for fit parameters or derived results
    * all derived results whose values and uncertainties are the same for all
      repetitions are aggregated together to give a single result. These results have
      the same name as the original model's results, with no suffix.
    * all derived results whose value is not the same for al repetitions are omitted.
    """

    def __init__(
        self,
        model: Model,
        common_params: Optional[List[str]] = None,
        num_repetitions: int = 2,
        aggregate_results=False,
    ):
        """
        :param model: The :class:`~ionics_fits.common.Model` to be repeated. The
          implementation of ``model`` will be used to generate data for all y axes. This
          model is considered owned by the ``RepeatedModel`` and should not be used /
          modified elsewhere.
        :param common_params: optional list of names of model parameters, whose value is
          common to all y axes. All other model parameters are allowed to vary
          independently between the y axes
        :param num_repetitions: the number of times the model is to be repeated
        :param aggregate_results: determines whether derived results are aggregated or
          not. The default behaviour is to not aggregate results. This is
          generally suitable when one wants access to the values of non-common
          parameters from the various repetitions. Aggregating results can be useful,
          for example, when all parameters are common across the repetitions and one
          wants a single set of values reported.
        """
        self.aggregate_results = aggregate_results
        model_params = set(model.parameters.keys())
        common_params = set(common_params or [])

        if not common_params.issubset(model_params):
            raise ValueError(
                "Common parameters must be a subset of the model's parameters"
            )

        independent_params = set(model_params) - common_params
        params = {param: model.parameters[param] for param in common_params}
        for param in independent_params:
            params.update(
                {
                    f"{param}_{idx}": param_like(model.parameters[param])
                    for idx in range(num_repetitions)
                }
            )

        self.model = model
        self.common_params = common_params
        self.independent_params = independent_params
        self.num_repetitions = num_repetitions

        super().__init__(
            parameters=params, internal_parameters=self.model.internal_parameters
        )

        def wrapped_scale_func(
            scale_func: Callable, x_scales: TX_SCALE, y_scales: TY_SCALE
        ) -> float:
            return scale_func(x_scales, y_scales[0 : self.model.get_num_y_axes()])

        parameters = list(self.parameters.values()) + self.internal_parameters
        for parameter in parameters:
            scale_func = parameter.scale_func
            parameter.scale_func = partial(wrapped_scale_func, scale_func)
            parameter.scale_func.__name__ = scale_func.__name__

    def get_num_x_axes(self) -> int:
        return self.model.get_num_x_axes()

    def get_num_y_axes(self) -> int:
        return self.num_repetitions * self.model.get_num_y_axes()

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        # Disable y-axis rescaling so we don't end up with multiple different scale
        # factors for the common parameters. To support rescaling here we would need to
        # extend Fitter to allow the model to have some control over the axis scale
        # factors.
        x_scales, _ = self.model.can_rescale()
        return x_scales, [False] * self.get_num_y_axes()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        x = np.atleast_2d(x)
        values = {param: param_values[param] for param in self.common_params}

        dim = self.model.get_num_y_axes()
        num_samples = x.shape[1]
        ys = np.zeros((self.get_num_y_axes(), num_samples))
        for idx in range(self.num_repetitions):
            values.update(
                {
                    param: param_values[f"{param}_{idx}"]
                    for param in self.independent_params
                }
            )
            ys[idx * dim : (idx + 1) * dim, :] = np.atleast_2d(
                self.model.func(x, values)
            )

        return ys

    def estimate_parameters(self, x: TX, y: TY):
        dim = self.model.get_num_y_axes()

        common_heuristics = {
            param: np.zeros(self.num_repetitions + 1) for param in self.common_params
        }

        for idx in range(self.num_repetitions):
            params = {
                param: self.parameters[f"{param}_{idx}"]
                for param in self.independent_params
            }

            params.update(
                {param: self.parameters[param] for param in self.common_params}
            )

            # Reset heuristics before each iteration
            for param_data in params.values():
                param_data.heuristic = None

            self.model.parameters = params
            self.model.estimate_parameters(x, y[idx * dim : (idx + 1) * dim])

            for param in self.common_params:
                common_heuristics[param][idx] = params[param].get_initial_value()

        # Combine the heuristics for the repetitions to find the best set of common
        # parameter values
        for param in self.common_params:
            common_heuristics[param][-1] = np.mean(common_heuristics[param][:-1])

        param_estimates = {
            param_name: self.parameters[param_name].get_initial_value()
            for param_name in self.parameters.keys()
            if param_name not in self.common_params
        }
        costs = np.zeros(self.num_repetitions + 1)
        for idx in range(self.num_repetitions + 1):
            y_idx = self.__call__(
                x=x,
                **param_estimates,
                **{
                    param_name: common_heuristics[param_name][idx]
                    for param_name in self.common_params
                },
            )
            costs[idx] = np.sqrt(np.sum((y - y_idx) ** 2))
        best_heuristic = np.argmin(costs)

        for param in self.common_params:
            self.parameters[param].heuristic = common_heuristics[param][best_heuristic]

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_uncertainties = {}

        dim = self.model.get_num_y_axes()

        # {param_name: [derived_values_from_each_repetition]}
        derived_params_reps = None
        derived_uncertainties_reps = None
        derived_param_names = None

        common_fitted_params = {
            param: fitted_params[param] for param in self.common_params
        }
        common_fit_uncertainties = {
            param: fit_uncertainties[param] for param in self.common_params
        }

        for idx in range(self.num_repetitions):
            rep_params = {
                param: fitted_params[f"{param}_{idx}"]
                for param in self.independent_params
            }
            rep_params.update(common_fitted_params)

            rep_uncertainties = {
                param: fit_uncertainties[f"{param}_{idx}"]
                for param in self.independent_params
            }
            rep_uncertainties.update(common_fit_uncertainties)

            derived = self.model.calculate_derived_params(
                x=x,
                y=y[idx * dim : (idx + 1) * dim],
                fitted_params=rep_params,
                fit_uncertainties=rep_uncertainties,
            )
            rep_derived_params, rep_derived_uncertainties = derived

            if derived_params_reps is None:
                derived_param_names = rep_derived_params.keys()
                # We can't preallocate prior to this because we don't know what the
                # model's derived parameter values are until we've calculated them
                derived_params_reps = {
                    param_name: [0.0] * self.num_repetitions
                    for param_name in derived_param_names
                }
                derived_uncertainties_reps = {
                    param_name: [0.0] * self.num_repetitions
                    for param_name in derived_param_names
                }

            for param_name in derived_param_names:
                derived_params_reps[param_name][idx] = rep_derived_params[param_name]
                derived_uncertainties_reps[param_name][idx] = rep_derived_uncertainties[
                    param_name
                ]

        def add_statistics(values, uncertainties, param_names):
            for param_name in param_names:
                param_values = np.array(
                    [
                        values[f"{param_name}_{idx}"]
                        for idx in range(self.num_repetitions)
                    ]
                )
                param_uncerts = np.array(
                    [
                        uncertainties[f"{param_name}_{idx}"]
                        for idx in range(self.num_repetitions)
                    ]
                )

                derived_params[f"{param_name}_mean"] = np.mean(param_values)
                derived_uncertainties[f"{param_name}_mean"] = (
                    np.sqrt(np.sum(param_uncerts**2)) / self.num_repetitions
                )

                derived_params[f"{param_name}_peak_peak"] = np.ptp(param_values)
                derived_uncertainties[f"{param_name}_peak_peak"] = float("nan")

        if self.aggregate_results:
            keep = [
                param_name
                for param_name in derived_param_names
                if np.all(
                    np.isclose(
                        derived_params_reps[param_name],
                        derived_params_reps[param_name][0],
                    )
                )
                and np.all(
                    np.isclose(
                        derived_uncertainties_reps[param_name],
                        derived_uncertainties_reps[param_name][0],
                    )
                )
            ]
            derived_params.update(
                {param_name: derived_params_reps[param_name][0] for param_name in keep}
            )
            derived_uncertainties.update(
                {
                    param_name: derived_uncertainties_reps[param_name][0]
                    for param_name in keep
                }
            )
        else:
            for param_name in derived_param_names:
                derived_params.update(
                    {
                        f"{param_name}_{idx}": derived_params_reps[param_name][idx]
                        for idx in range(self.num_repetitions)
                    }
                )
                derived_uncertainties.update(
                    {
                        f"{param_name}_{idx}": derived_uncertainties_reps[param_name][
                            idx
                        ]
                        for idx in range(self.num_repetitions)
                    }
                )

            add_statistics(
                derived_params,
                derived_uncertainties,
                param_names=rep_derived_params.keys(),
            )
            add_statistics(
                fitted_params,
                fit_uncertainties,
                param_names=self.independent_params,
            )

        return derived_params, derived_uncertainties



r""" It's not enough to just fit data, we want to know whether we can trust the fit
results. This is where validators come in.

There are two distinct aspects to the validation problem:

* did the fit find the model parameters which best match the data (as opposed to getting
  stuck in a local minimum in parameter space far from the global optimum)?
* Are the fitted parameter values consistent with our prior knowledge of the system
  (e.g. we know that a fringe contrast must lie within certain bounds).

The approach taken in ``ionics_fits`` is as follows. First any prior knowledge about the
system should be encoded into the :class:`~ionics_fits.common.ModelParameter`
configuration by setting parameter bounds, fixing parameters, etc.

After that, the fit is validated using a :class:`FitValidator`. Validators provide a
flexible and extensible framework for using statistical tests to validate fits.
"""

from typing import Tuple
import numpy as np


from .common import Fitter


class FitValidator:
    """Base class for fit validators"""

    def window_fun(self, fit_results: Fitter):
        """Returns a boolean array of x-axis points which should be included in
        fit validation. This implementation includes all x-axis data points;
        override to only include a subset.
        """
        return np.full(fit_results.x.shape[1], True)

    def validate(self, fit_results: Fitter) -> Tuple[bool, float]:
        """Validates the fit.

        Subclasses must override this method

        :returns: tuple specifying whether the fit succeeded and the fit significance
          as a number between 0 (complete failure) and 1 (perfect fit).
        """
        raise NotImplementedError


class NSigmaValidator(FitValidator):
    def __init__(self, n_sigma: float = 3.0, significance_threshold: float = 0.75):
        """Fit validator which checks that at least
        ``significance_threshold`` of points lie within ``n_sigma``
        standard errors of the fitted value.

        This is a relatively forgiving (easy to configure in a way that gives
        minimal "good" fits which fail validation) general-purpose fit validator.

        :param n_sigma: number of standard errors that points allowed to differ from the
          model.
        :param significance_threshold: fraction of points which must lie within
          ``n_sigma`` of the model for the fit to be considered successful.
        """
        self.n_sigma = n_sigma
        self.significance_threshold = significance_threshold

    def validate(self, fit_results: Fitter) -> Tuple[bool, float]:
        if fit_results.sigma is None:
            raise ValueError(
                "Cannot validate fit without standard errors for each point"
            )

        window = self.window_fun(fit_results)

        x = fit_results.x[:, window]
        y = fit_results.y[:, window]
        sigma = fit_results.sigma[:, window]

        _, y_fit = fit_results.evaluate(x_fit=x)

        errs = np.abs((y - y_fit) / sigma)
        errs = np.sort(errs.ravel())

        good_points = errs < self.n_sigma
        good_fraction = sum(good_points) / len(good_points)

        return good_fraction >= self.significance_threshold, good_fraction



import numpy as np
from typing import Callable, List, Union, TYPE_CHECKING


if TYPE_CHECKING:
    num_x_axes = float
    num_y_axes = float


class _SubscriptableNumpyArray(type):
    def __getitem__(cls, _):
        return np.ndarray


class Array(np.ndarray, metaclass=_SubscriptableNumpyArray):
    """Subclass of numpy's NDArray used purely for type annotation convenience.

    Type annotations can have arbitrary subscripts, e.g. ``Array[(4,), "float64"]``.
    """


class _SubscriptableArrayLikeUnion(type):
    def __getitem__(cls, x):
        return Union[List, Array[x]]


class ArrayLike(metaclass=_SubscriptableArrayLikeUnion):
    """
    Used for types that can be a numpy array or a list that's then converted to an array
    """


TX_SCALE = Array[("num_x_axes",), np.float64]
TY_SCALE = Array[("num_y_axes",), np.float64]
TSCALE_FUN = Callable[[TX_SCALE, TY_SCALE], float]


def scale_invariant(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
    r"""Scale function for :class:`~ionics_fits.common.ModelParameter`\ s whose value is
    invariant under rescaling of the x- and y-axes

    :param x_scales: array of x-axis scale factors
    :param y_scales: array of y-axis scale factors
    :returns: ``1``
    """
    return 1


def scale_x(x_axis: int = 0) -> TSCALE_FUN:
    r"""Returns a scale function for :class:`~ionics_fits.common.ModelParameter`\ s whose
    value scales linearly with one x-axis dimension.

    :param x_axis: index of the x-axis dimension the parameter scales with
    :returns: scale function
    """

    def fun(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
        return x_scales[x_axis]

    fun.__name__ = "scale_x"
    return fun


def scale_x_inv(x_axis: int = 0) -> TSCALE_FUN:
    r"""Returns a scale function for :class:`~ionics_fits.common.ModelParameter`\ s whose
    value scales inversely with one x-axis dimension.

    :param x_axis: index of the x-axis dimension the parameter scales with
    :returns: scale function
    """

    def fun(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
        return 1 / x_scales[x_axis]

    fun.__name__ = "scale_x_inv"
    return fun


def scale_y(y_axis: int = 0) -> TSCALE_FUN:
    r"""Returns a scale function for :class:`~ionics_fits.common.ModelParameter`\ s whose
    value scales linearly with one y-axis dimension.

    :param y_axis: index of the y-axis dimension the parameter scales with
    :returns: scale function
    """

    def fun(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
        return y_scales[y_axis]

    fun.__name__ = "scale_y"
    return fun


def scale_power(
    x_power: int, y_power: int, x_axis: int = 0, y_axis: int = 0
) -> TSCALE_FUN:
    r"""Returns a scale function for :class:`~ionics_fits.common.ModelParameter`\ s whose
    value scales as a function of one x-axis and one y-axis dimension.

    The parameter scale factor is calculated as::

        scale_factor = (x_scales[x_axis] ** x_power) * (y_scales[y_axis] ** y_power)

    :param x_power: x-axis power
    :param y_power: y-axis power
    :param x_axis: index of the x-axis dimension the parameter scales with
    :param y_axis: index of the y-axis dimension the parameter scales with
    :returns: scale function
    """

    def fun(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
        return (x_scales[x_axis] ** x_power) * (y_scales[y_axis] ** y_power)

    fun.__name__ = "scale_power"
    return fun


def scale_undefined(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
    r"""Scale function for :class:`~ionics_fits.common.ModelParameter`\ s whose scaling
    is not known yet.

    This scale function is typically used for parameters whose scale factor is not known
    until runtime.

    :param x_scales: array of x-axis scale factors
    :param y_scales: array of y-axis scale factors
    """

    raise RuntimeError(
        "Attempt to rescale model parameter with undefined scale function"
    )


def scale_no_rescale(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
    r"""Scale function for :class:`~ionics_fits.common.ModelParameter`\ s which cannot be
    rescaled.

    Raises a ``RuntimeError`` if any of the x-axis or y-axis scale factors are not equal
    to ``1``.

    :param x_scales: array of x-axis scale factors
    :param y_scales: array of y-axis scale factors
    :returns: ``1``.
    """
    if any([x_scale != 1.0 for x_scale in x_scales]):
        raise RuntimeError(
            "Attempt to rescale model parameter along x which does not support "
            "rescaling"
        )
    if any([y_scale != 1.0 for y_scale in y_scales]):
        raise RuntimeError(
            "Attempt to rescale model parameter along y which does not support "
            "rescaling"
        )

    return 1.0



from typing import Dict, List, Tuple
import numpy as np

from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_x, scale_y


class Exponential(Model):
    """Exponential function according to::

        y(x < x_dead) = y0
        y(x >= x_dead) = y0 + (y_inf-y0)*(1-exp(-(x-x_dead)/tau))

    See :meth:`_func` for parameter details.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x_dead: ModelParameter(
            lower_bound=0,
            fixed_to=0,
            scale_func=scale_x(),
        ),
        y0: ModelParameter(scale_func=scale_y()),
        y_inf: ModelParameter(scale_func=scale_y()),
        tau: ModelParameter(lower_bound=0, scale_func=scale_x()),
    ) -> TY:
        """
        :param x_dead: x-axis "dead time" (fixed to ``0`` by default)
        :param y0: initial (``x = x_dead``) y-axis offset
        :param y_inf: y-axis asymptote (i.e. ``y(x - x_dead >> tau) => y_inf``)
        :param tau: decay constant
        """
        y = y0 + (y_inf - y0) * (1 - np.exp(-(x - x_dead) / tau))
        y = np.where(x >= x_dead, y, y0)
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        # Ensure that y is a 1D array
        x = np.squeeze(x)
        y = np.squeeze(y)

        # Exponentials are generally pretty easy to fit so we keep the estimator simple
        self.parameters["x_dead"].heuristic = 0
        self.parameters["y0"].heuristic = y[0]
        self.parameters["y_inf"].heuristic = y[-1]
        self.parameters["tau"].heuristic = x.ptp()

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Derived parameters:

            * ``x_1_e``: x-axis value for ``1/e`` decay including dead time
              (``x_1_e = x_dead + tau``)

        """
        derived_params = {"x_1_e": fitted_params["x_dead"] + fitted_params["tau"]}
        derived_uncertainties = {
            "x_1_e": np.sqrt(
                fit_uncertainties["x_dead"] ** 2 + fit_uncertainties["tau"] ** 2
            )
        }
        return derived_params, derived_uncertainties



import copy
import dataclasses
from typing import Any, Dict, List, Optional, TYPE_CHECKING

from ..common import ModelParameter, TSCALE_FUN


if TYPE_CHECKING:
    num_samples = float
    num_spectrum_samples = float


def param_like(
    template_param: ModelParameter, overrides: Optional[Dict[str, Any]] = None
) -> ModelParameter:
    """Returns a new parameter based on a template.

    :param template_param: the returned parameter is a (deep) copy of the template
      parameter.
    :param overrides: optional dictionary of attributes of the template parameter to
      replace.
    :returns: the new :class:`~ionics_fits.common.ModelParameter`
    """
    new_param = copy.deepcopy(template_param)

    overrides = overrides or {}
    for attr, value in overrides.items():
        setattr(new_param, attr, value)

    return new_param


@dataclasses.dataclass
class PeriodicModelParameter(ModelParameter):
    """Represents a model parameter whose value is periodic.

    Parameter values are clipped to lie within::

        value = value - offset
        value (value % period) + offset

    Attributes:
        period: the period (default = 1)
        offset: the offset (default = 0)
    """

    scale_func: TSCALE_FUN
    period: float = 1
    offset: float = 0
    lower_bound: float = dataclasses.field(init=False)
    upper_bound: float = dataclasses.field(init=False)

    def __post_init__(self):
        super().__post_init__()
        self._metadata_attrs += ["period", "offset"]

        self.lower_bound = 1.5 * self.offset
        self.upper_bound = 1.5 * (self.offset + self.period)

    def _format_metadata(self) -> List[str]:
        metadata = super()._format_metadata()

        metadata = [
            attr
            for attr in metadata
            if not (attr.startswith("lower_bound=") or attr.startswith("upper_bound="))
        ]

        metadata.append(f"period={self.period:.3f}")
        metadata.append(f"offset={self.offset:.3f}")

        return metadata

    def __repr__(self):
        return super().__repr__()

    def clip(self, value: float):
        """Clip value to lie between lower and upper bounds."""
        value = value - self.offset
        return (value % self.period) + self.offset



from typing import List, Tuple
import numpy as np

from . import heuristics
from .rectangle import Rectangle
from .triangle import Triangle
from ..common import Model, ModelParameter, TX, TY
from ..normal import NormalFitter
from ..utils import scale_x, scale_x_inv, scale_y


class Sinc(Model):
    """Sinc function according to::

        y = a * sin(w * (x - x0)) / (w * (x - x0)) + y0

    See :meth:`_func` for parameters.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x0: ModelParameter(scale_func=scale_x()),
        y0: ModelParameter(scale_func=scale_y()),
        a: ModelParameter(scale_func=scale_y()),
        w: ModelParameter(lower_bound=0, scale_func=scale_x_inv()),
    ) -> TY:
        """
        :param x0: x-axis offset
        :param y0: y-axis offset
        :param a: amplitude
        :param w: x scale factor
        """
        x = w * (x - x0) / np.pi  # np.sinc(x) = sin(pi*x) / (pi*x)
        y = a * np.sinc(x) + y0
        return y

    # pytype: enable=invalid-annotation
    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        self.parameters["y0"].heuristic = np.mean([y[0], y[-1]])
        y0 = self.parameters["y0"].get_initial_value()

        omega, spectrum = heuristics.get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        # Fourier transform of a sinc is a rectangle
        rect = Rectangle()
        rect.parameters["y0"].fixed_to = 0
        rect.parameters["x_l"].fixed_to = 0
        rect.parameters["a"].heuristic = max(abs_spectrum)

        fit = NormalFitter(omega, abs_spectrum, model=rect)

        self.parameters["w"].heuristic = fit.values["x_r"]
        w = self.parameters["w"].get_initial_value() + abs_spectrum[1]

        sgn = 1 if y[np.argmax(np.abs(y - y0))] > y0 else -1
        self.parameters["a"].heuristic = (w / np.pi) * fit.values["a"] * sgn

        x0 = heuristics.find_x_offset_sym_peak_fft(
            model=self,
            x=x,
            y=y,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=w,
        )

        self.parameters["x0"].heuristic = x0


class Sinc2(Model):
    """Sinc-squared function according to::

        y = a * (sin(w * (x - x0)) / (w * (x - x0)))^2 + y0

    See :meth:`_func` for parameters.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x0: ModelParameter(scale_func=scale_x()),
        y0: ModelParameter(scale_func=scale_y()),
        a: ModelParameter(scale_func=scale_y()),
        w: ModelParameter(lower_bound=0, scale_func=scale_x_inv()),
    ) -> TY:
        """
        :param x0: x-axis offset
        :param y0: y-axis offset
        :param a: amplitude
        :param w: x scale factor
        """
        x = w * (x - x0) / np.pi  # np.sinc(x) = sin(pi*x) / (pi*x)
        y = a * (np.sinc(x) ** 2) + y0
        return y

    # pytype: enable=invalid-annotation
    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        y0 = self.parameters["y0"].heuristic = np.mean([y[0], y[-1]])

        omega, spectrum = heuristics.get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        # Fourier transform of a sinc^2 is a triangle function
        tri = Triangle()
        tri.parameters["x0"].fixed_to = 0
        tri.parameters["y0"].heuristic = max(abs_spectrum)
        tri.parameters["sym"].fixed_to = 0
        tri.parameters["y_min"].fixed_to = 0

        fit = NormalFitter(omega, abs_spectrum, model=tri)

        intercept = fit.values["y0"] / -fit.values["k"]
        sgn = 1 if y[np.argmax(np.abs(y - y0))] > y0 else -1

        w = self.parameters["w"].get_initial_value(intercept / 2)
        self.parameters["w"].heuristic = w
        self.parameters["a"].heuristic = (
            fit.values["y0"] * sgn * intercept / (2 * np.pi)
        )

        x0 = heuristics.find_x_offset_sym_peak_fft(
            model=self,
            x=x,
            y=y,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=intercept,
        )

        self.parameters["x0"].heuristic = x0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ionics_fits', 'ionics_fits.models', 'ionics_fits.models.transformations']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.2,<2.0.0', 'scipy>=1.7.1,<2.0.0', 'statsmodels>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'ionics-fits',
    'version': '1.3.8',
    'description': 'Lightweight Python data fitting library with an emphasis on AMO/Quantum Information',
    'long_description': 'Lightweight Python library for data fitting with an emphasis on AMO (Atomic Molecular\nand Optical physics) and Quantum Information.\n\nSee the documentation at: https://oxionics.github.io/ionics_fits/\n',
    'author': 'hartytp',
    'author_email': 'thomas.peter.harty@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
Models with more than one x-axis degree of freedom, which have been created from 1D
models using :class:`~ionics_fits.models.transformations.model_2d.Model2D`.
"""
import numpy as np

from .cone import ConeSlice
from .gaussian import Gaussian
from .polynomial import Parabola
from .transformations.mapped_model import MappedModel
from .transformations.model_2d import Model2D
from .triangle import Triangle

from ..utils import TX_SCALE, TY_SCALE


class Gaussian2D(MappedModel):
    """2D Gaussian according to::

        y = (
            a / ((sigma_x0 * sqrt(2*pi)) * (sigma_x1 * sqrt(2*pi)))
            * exp(-0.5*((x0-x0_x0)/(sigma_x0))^2 -0.5*((x1-x0_x1)/(sigma_x1))^2) + y0

    Parameters are:

    * ``a``
    * ``x0_x0``
    * ``x0_x1``
    * ``sigma_x0``
    * ``sigma_x1``
    * ``y0``

    Derived results are:
      - ``FWHMH_x0``
      - ``FWHMH_x1``
      - ``w0_x0``
      - ``w0_x1``
      - ``peak``

    See :class:`~ionics_fits.models.gaussian.Gaussian` for details.
    """

    def __init__(self):
        class _Gaussian2D(Model2D):
            def __init__(self):
                super().__init__(
                    models=(Gaussian(), Gaussian()),
                    result_params=("a",),
                )

            def wrap_scale_funcs(self):
                super().wrap_scale_funcs()

                def scale_power(x_scales: TX_SCALE, y_scales: TY_SCALE) -> float:
                    return np.prod(x_scales) * float(y_scales)

                self.parameters["a_x1"].scale_func = scale_power

        model = _Gaussian2D()

        param_mapping = {
            param_name: param_name
            for param_name in model.parameters.keys()
            if param_name not in ["y0_x0", "y0_x1", "a_x1"]
        }
        param_mapping["y0"] = "y0_x0"
        param_mapping["a"] = "a_x1"
        super().__init__(
            model=model,
            param_mapping=param_mapping,
            fixed_params={"y0_x1": 0},
            derived_result_mapping={"peak": "peak_x1", None: "peak_x0"},
        )


class Parabola2D(MappedModel):
    """2D Parabola according to::

        y = k_x0 * (x0 - x0_x0)^2 + k_x1 *(x1 - x0_x1) + y0


    Parameters are:

    * ``x0_x0``
    * ``x0_x1``
    * ``k_x0``
    * ``k_x1``
    * ``y0``

    See :class:`~ionics_fits.models.polynomial.Parabola` for details.
    """

    def __init__(self):
        model = Model2D(models=(Parabola(), Parabola()), result_params=("y0",))
        param_mapping = {
            param_name: param_name
            for param_name in model.parameters.keys()
            if param_name != "y0_x1"
        }
        param_mapping["y0"] = "y0_x1"
        super().__init__(model=model, param_mapping=param_mapping)


class Cone2D(MappedModel):
    r"""2D Cone Model.

    Parameters are:

    - ``x0_x0``
    - ``x0_x1``
    - ``k_x0``
    - ``k_x1``
    - ``y0``

    Parameters with an ``_x0`` suffix inherit from
    :class:`~ionics_fits.models.cone.ConeSlice`\, parameters with an ``_x1`` suffix
    inherit from :class:`~ionics_fits.models.triangle.Triangle`.
    """

    def __init__(self):
        model = Model2D(models=(ConeSlice(), Triangle()), result_params=("alpha",))
        param_mapping = {
            param_name: param_name
            for param_name in model.parameters.keys()
            if param_name != "z0_x0"
        }

        param_mapping["y0"] = "z0_x0"

        super().__init__(model=model, param_mapping=param_mapping)



from typing import Dict, List, Tuple

import numpy as np

from . import heuristics
from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_invariant, scale_undefined, scale_x, scale_x_inv


class MolmerSorensen(Model):
    r"""Base class for Mølmer–Sørensen interactions.

    This model calculates the time-dependent populations for one or two qubits
    coupled to a single motional mode undergoing a Mølmer–Sørensen type
    interaction.

    It requires that the initial spin states of all qubits are the same
    and either ``|g>`` or ``|e>`` - different initial states for each qubit or initial
    states which are superpositions of spin eigenstates are are not supported.

    For single-qubit interactions, the model has one y-axis dimension, giving the
    excited-state population at the end of the interaction.

    For two-qubit interactions, the model has three y-axis dimensions - ``P_gg``\,
    ``P_1e``\, ``P_ee`` - giving the probabilities of 0, 1 or 2 ions being in the
    excited state at the end of the interaction duration.

    Modulation of the sign of the spin-dependent force according to a Walsh
    function is supported.

    The motion's initial state must be a thermal distribution.

    This class does not support fitting directly; use one of its subclasses
    instead.

    Independent variables:

        * ``t_pulse``: total interaction duration.
        * ``w``: detuning of red/blue sideband tones relative to reference frequency
          ``w_0``. The interaction detuning is given by ``delta = w - w_0``.

    All frequencies are in angular units unless stated otherwise.
    """

    def __init__(self, num_qubits: int, start_excited: bool, walsh_idx: int):
        r"""
        :param num_qubits: number of qubits (must be 1 or 2)
        :param walsh_idx: Index of Walsh function
        :param start_excited: If True, all qubits start in ``|e>``\ , otherwise they
            start in ``|g>``.
        """
        super().__init__()

        if num_qubits not in [1, 2]:
            raise ValueError("Model only supports 1 or 2 qubits")
        if walsh_idx not in [0, 1, 2, 3]:
            raise ValueError("Unsupported Walsh index")

        self.num_qubits = num_qubits
        self.start_excited = start_excited
        self.walsh_idx = walsh_idx

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return [1, 3][self.num_qubits - 1]

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [False] * self.get_num_y_axes()

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        omega: ModelParameter(lower_bound=0.0, scale_func=scale_undefined),
        w_0: ModelParameter(scale_func=scale_undefined),
        n_bar: ModelParameter(
            lower_bound=0.0, fixed_to=0.0, scale_func=scale_invariant
        ),
    ) -> TY:
        r"""Return measurement probabilities for the states ``P_gg``\, ``P_1``\, and
        ``P_ee``.

        :param omega: sideband Rabi frequency
        :param w_0: angular resonance frequency offset
        :param n_bar: average initial occupancy of the motional mode
        """

        t_pulse = x[0]
        delta = x[1] - w_0
        data_shape = np.broadcast(t_pulse, delta).shape

        if self.walsh_idx == 0:
            segment_durations = [1]
        elif self.walsh_idx == 1:
            segment_durations = [1, 1]
        elif self.walsh_idx == 2:
            segment_durations = [1, 1, 1, 1]
        else:
            segment_durations = [1, 2, 1]
        segment_durations = np.array(segment_durations, dtype=float)
        segment_durations /= sum(segment_durations)

        def displacement(delta, t_i, t_f):
            alpha = np.full(shape=data_shape, fill_value=-1j * t_pulse)
            alpha = (
                0.5
                * omega
                * np.divide(
                    np.exp(-1j * delta * t_f) - np.exp(-1j * delta * t_i),
                    delta,
                    out=alpha,
                    where=(delta != 0),
                )
            )
            return alpha

        alpha = np.zeros(data_shape, dtype=np.complex128)
        t_i = 0
        sign = +1
        for duration in segment_durations:
            t_f = t_i + duration * t_pulse
            alpha += sign * displacement(delta=delta, t_i=t_i, t_f=t_f)
            t_i = t_f
            sign *= -1

        gamma = np.abs(alpha) ** 2 * (n_bar + 1 / 2)

        if self.num_qubits == 1:
            P = 0.5 * (1 + np.exp(-4 * gamma))
            if self.start_excited:
                return P
            else:
                return 1 - P

        # piecewise integration of the geometric phase
        phi = delta * t_pulse
        t_outer_i = 0
        for outer_idx, outer_duration in enumerate(segment_durations):
            t_outer_f = t_outer_i + outer_duration * t_pulse

            inner_segments = segment_durations[0:outer_idx]
            t_inner_i = 0
            for inner_idx, inner_duration in enumerate(inner_segments):
                t_inner_f = t_inner_i + inner_duration * t_pulse
                sign = (-1) ** (outer_idx - inner_idx)
                phi += sign * (
                    np.sin(delta * (t_outer_f - t_inner_f))
                    - np.sin(delta * (t_outer_f - t_inner_i))
                    - np.sin(delta * (t_outer_i - t_inner_f))
                    + np.sin(delta * (t_outer_i - t_inner_i))
                )
                t_inner_i = t_inner_f

            phi -= np.sin(delta * (t_outer_f - t_inner_i))
            t_outer_i = t_outer_f

        phi *= np.square(
            np.divide(omega, 2 * delta, out=np.zeros(data_shape), where=(delta != 0))
        )

        # Calculate spin transition probabilities
        P_0 = (
            1 / 8 * (3 + np.exp(-16 * gamma) + 4 * np.cos(4 * phi) * np.exp(-4 * gamma))
        )

        P_1 = 1 / 4 * (1 - np.exp(-16 * gamma))

        P_2 = (
            1 / 8 * (3 + np.exp(-16 * gamma) - 4 * np.cos(4 * phi) * np.exp(-4 * gamma))
        )

        if self.start_excited:
            P_gg = P_2
            P_ee = P_0
        else:
            P_gg = P_0
            P_ee = P_2

        return np.vstack([P_gg, P_1, P_ee])

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Derived parameters:

        * ``f_0``: resonance frequency offset (Hz)
        """
        derived_params = {}
        w_0_param = "w_0" if "w_0" in fitted_params.keys() else "delta"
        derived_params["f_0"] = fitted_params[w_0_param] / (2 * np.pi)

        derived_uncertainties = {}
        derived_uncertainties["f_0"] = fit_uncertainties[w_0_param] / (2 * np.pi)

        return derived_params, derived_uncertainties

    # pytype: enable=invalid-annotation


class MolmerSorensenTime(MolmerSorensen):
    r"""
    Fit model for Mølmer–Sørensen pulse duration scans.

    This model calculates the populations for Mølmer–Sørensen interactions
    when the interaction detuning is kept fixed and only the pulse duration
    is varied.

    Since the detuning is not scanned as an independent variable, we replace
    ``w_0`` with a new model parameter ``delta``\ , defined by ``delta = |w - w_0|``.
    """

    def __init__(self, num_qubits: int, start_excited: bool, walsh_idx: int):
        super().__init__(
            num_qubits=num_qubits, start_excited=start_excited, walsh_idx=walsh_idx
        )

        self.parameters["delta"] = ModelParameter(scale_func=scale_x_inv())
        del self.parameters["w_0"]

        self.parameters["omega"].scale_func = scale_x_inv()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        """
        Return populations as function of pulse duration for fixed detuning.

        :param x: pulse duration
        """
        param_values = param_values.copy()
        delta = param_values.pop("delta")
        param_values["w_0"] = 0.0
        return self._func((x, delta), **param_values)  # pytype: disable=wrong-arg-types

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)

        self.parameters["n_bar"].heuristic = 0.0
        if (
            not self.parameters["delta"].has_user_initial_value()
            and not self.parameters["omega"].has_user_initial_value()
        ):
            # This case is pretty miserable because there is a very high degree
            # of covariance between delta and omega so even if the heuristics are
            # highly accurate, the optimizer will tend to struggle to converge on
            # the right parameter values.
            self.parameters["delta"].heuristic = 0.0

        if not self.parameters["omega"].has_user_initial_value():
            omegas = np.array([np.linspace(0.1, 10, 25)]) * np.pi / max(x)
            self.parameters["omega"].heuristic, _ = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="omega",
                scanned_param_values=omegas,
            )

        elif not self.parameters["delta"].has_user_initial_value():
            omega = self.parameters["omega"].get_initial_value()
            deltas = np.array([np.linspace(0, 10, 25)]) * omega
            self.parameters["delta"].heuristic, _ = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="delta",
                scanned_param_values=deltas,
            )


class MolmerSorensenFreq(MolmerSorensen):
    r"""
    Fit model for Mølmer–Sørensen detuning scans.

    This model calculates the populations for Mølmer–Sørensen interactions
    when the gate duration is kept fixed and only the interaction detuning is
    varied. The pulse duration is specified using a new ``t_pulse`` model
    parameter.
    """

    def __init__(self, num_qubits: int, start_excited: bool, walsh_idx: int):
        super().__init__(
            num_qubits=num_qubits, start_excited=start_excited, walsh_idx=walsh_idx
        )

        self.parameters["t_pulse"] = ModelParameter(
            lower_bound=0.0, scale_func=scale_x_inv()
        )

        self.parameters["omega"].scale_func = scale_x()
        self.parameters["w_0"].scale_func = scale_x()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        """
        Return populations as function of detuning for fixed duration.

        :param x: Interaction detuning
        """
        param_values = param_values.copy()
        t_pulse = param_values.pop("t_pulse")
        return self._func(
            (t_pulse, x), **param_values
        )  # pytype: disable=wrong-arg-types

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        self.parameters["n_bar"].heuristic = 0.0

        # estimate the centre frequency by looking for the symmetry point.
        # Consider only the 1-ion transition probability
        y_test = y if self.num_qubits == 1 else np.atleast_2d(y[1, :])
        w_0 = self.parameters["w_0"].heuristic = heuristics.get_sym_x(x, y_test)

        if (
            self.parameters["t_pulse"].has_user_initial_value()
            and self.parameters["omega"].has_user_initial_value()
        ):
            if not self.parameters["w_0"].has_user_initial_value():
                # The symmetry heuristic is pretty good, but can struggle when
                # the centre frequency is close to the edge of the scan range.
                # Since w_0 is our only unknown here, we throw in a sampling
                # heuristic for good measure.
                w_0_grid, w_0_grid_cost = heuristics.param_min_sqrs(
                    model=self,
                    x=x,
                    y=y,
                    scanned_param="w_0",
                    scanned_param_values=np.linspace(min(x), max(x), 50),
                )
                param_values = {
                    name: param.get_initial_value()
                    for name, param in self.parameters.items()
                }
                y_sym = self.func(x, param_values)
                w_0_sym_cost = np.sqrt(np.sum(np.square(y - y_sym)))
                w_0 = w_0 if w_0_sym_cost < w_0_grid_cost else w_0_grid
                self.parameters["w_0"].heuristic = w_0
            return

        if self.parameters["t_pulse"].has_user_initial_value():
            t_pulse = self.parameters["t_pulse"].get_initial_value()
            self.parameters["omega"].heuristic, _ = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="omega",
                scanned_param_values=np.array([np.linspace(0.25, 5, 10)])
                * np.pi
                / t_pulse,
            )
        elif self.parameters["omega"].has_user_initial_value():
            omega = self.parameters["omega"].get_initial_value()
            self.parameters["t_pulse"].heuristic, _ = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="t_pulse",
                scanned_param_values=np.array([np.linspace(0.25, 5, 10)])
                * np.pi
                / omega,
            )
        else:
            # this is a bit of a corner case, since the user can usually give us
            # an estimate of one of omega or t_pulse. In the absence of that we
            # fall back to a 2D grid search.
            omegas = np.arange(start=1, stop=25) / 100 * max(abs(x))
            costs = np.zeros_like(omegas)
            t_pulses = np.zeros_like(omegas)
            for idx, omega in np.ndenumerate(omegas):
                self.parameters["omega"].heuristic = omega
                t_pulses[idx], costs[idx] = heuristics.param_min_sqrs(
                    model=self,
                    x=x,
                    y=y,
                    scanned_param="t_pulse",
                    scanned_param_values=np.array([np.linspace(0.1, 5, 20)])
                    * np.pi
                    / omega,
                )
            best = np.argmin(costs)
            self.parameters["omega"].heuristic = omegas[best]
            self.parameters["t_pulse"].heuristic = t_pulses[best]

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        r"""Derived parameters:

        * ``f_0``: resonance frequency offset (Hz)
        * ``f_loop_{n}_{i}``: frequency offset of ``n``\th loop closure (Hz) for
          ``n = [1, 5]`` at "plus" (``i = p``\) or "minus" (``i = m``\) detuning
        """

        derived_params, derived_uncertainties = super().calculate_derived_params(
            x, y, fitted_params, fit_uncertainties
        )

        for n in range(1, 6):
            derived_params[f"f_loop_{n}_p"] = (
                n / fitted_params["t_pulse"] + derived_params["f_0"]
            )
            derived_params[f"f_loop_{n}_m"] = (
                -n / fitted_params["t_pulse"] + derived_params["f_0"]
            )
            derived_uncertainties[f"f_loop_{n}_p"] = np.sqrt(
                (n * fit_uncertainties["t_pulse"]) ** 2
                + derived_uncertainties["f_0"] ** 2
            )
            derived_uncertainties[f"f_loop_{n}_m"] = np.sqrt(
                (n * fit_uncertainties["t_pulse"]) ** 2
                + derived_uncertainties["f_0"] ** 2
            )
        return derived_params, derived_uncertainties



from typing import Dict, List, Tuple
import numpy as np

from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_invariant, scale_power, scale_x, scale_y


class Triangle(Model):
    """Triangle function according to::

        y(x>=x0) = k_p*|x-x0| + y0
        y(x<x0) = k_m*|x-x0| + y0
        y = max(y, y_min)
        y = min(y, m_max)
        k_p = (1 + sym) * k
        k_m = (1 - sym) * k

    See :meth:`_func` for parameter details.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x0: ModelParameter(scale_func=scale_x()),
        y0: ModelParameter(scale_func=scale_y()),
        k: ModelParameter(scale_func=scale_power(x_power=-1, y_power=1)),
        sym: ModelParameter(
            lower_bound=-1, upper_bound=1, fixed_to=0, scale_func=scale_invariant
        ),
        y_min: ModelParameter(fixed_to=-np.inf, scale_func=scale_y()),
        y_max: ModelParameter(fixed_to=+np.inf, scale_func=scale_y()),
    ) -> TY:
        """
        :param x0: x-axis offset
        :param y0: y-axis offset
        :param k: average slope
        :param sym: symmetry parameter (fixed to ``0`` by default)
        :param y_min: minimum value of y (bound to ``-inf`` by default)
        :param y_max: maximum value of y (bound to ``+inf`` by default)
        """
        k_p = k * (1 + sym)
        k_m = k * (1 - sym)

        y_p = k_p * np.abs(x - x0) + y0
        y_m = k_m * np.abs(x - x0) + y0
        y = np.where(x >= x0, y_p, y_m)
        y = np.where(y > y_min, y, y_min)
        y = np.where(y < y_max, y, y_max)

        return y

    # pytype: enable=invalid-annotation
    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        # Written to handle the case of data which is only well-modelled by a
        # triangle function near `x0` but saturates further away
        self.parameters["y_max"].heuristic = max(y)
        self.parameters["y_min"].heuristic = min(y)

        min_ind = np.argmin(y)
        max_ind = np.argmax(y)

        y_min = y[min_ind]
        y_max = y[max_ind]
        x_min = x[min_ind]
        x_max = x[max_ind]

        # Case 1: positive slope with peaks left and right of x_min above y_min
        x_l = x[x <= x_min]
        x_r = x[x >= x_min]
        y_l = y[x <= x_min]
        y_r = y[x >= x_min]

        left_peak_ind = np.argmax(y_l)
        right_peak_ind = np.argmax(y_r)

        dx_l = x_min - x_l[left_peak_ind]
        dx_r = x_r[right_peak_ind] - x_min

        k_m = (y_l[left_peak_ind] - y_min) / dx_l if dx_l != 0 else 0
        k_p = (y_r[right_peak_ind] - y_min) / dx_r if dx_r != 0 else 0
        alpha = 0 if k_m == 0 else k_p / k_m

        positive_defaults = {
            "x0": x_min,
            "y0": y_min,
            "k": 0.5 * (k_p + k_m),
            "sym": (alpha - 1) / (1 + alpha),
        }
        positive_parameters = {
            param: param_data.get_initial_value(default=positive_defaults.get(param))
            for param, param_data in self.parameters.items()
        }
        positive_cost = np.sum((y - self._func(x, **positive_parameters)) ** 2)

        # Case 2: negative slope with peaks left and right of x_max below y_max
        x_l = x[x <= x_max]
        x_r = x[x >= x_max]
        y_l = y[x <= x_max]
        y_r = y[x >= x_max]

        left_peak_ind = np.argmin(y_l)
        right_peak_ind = np.argmin(y_r)

        dx_l = x_max - x_l[left_peak_ind]
        dx_r = x_r[right_peak_ind] - x_max

        k_m = (y_l[left_peak_ind] - y_max) / dx_l if dx_l != 0 else 0
        k_p = (y_r[right_peak_ind] - y_max) / dx_r if dx_r != 0 else 0
        alpha = 0 if k_m == 0 else k_p / k_m

        negative_defaults = {
            "x0": x_max,
            "y0": y_max,
            "k": 0.5 * (k_p + k_m),
            "sym": (alpha - 1) / (1 + alpha),
        }
        negative_parameters = {
            param: param_data.get_initial_value(default=negative_defaults.get(param))
            for param, param_data in self.parameters.items()
        }
        negative_cost = np.sum((y - self._func(x, **negative_parameters)) ** 2)

        if positive_cost < negative_cost:
            best_params = positive_parameters
        else:
            best_params = negative_parameters

        for param, value in best_params.items():
            self.parameters[param].heuristic = value

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Derived parameters:

        * k_m: slope for ``x < x0``
        * k_p: slope for ``x >= x0``
        """
        derived_params = {}
        k = fitted_params["k"]
        sym = fitted_params["sym"]

        k_err = fitted_params["k"]
        sym_err = fitted_params["sym"]

        derived_params["k_p"] = (1 + sym) * k
        derived_params["k_m"] = (1 - sym) * k

        derived_uncertainties = {}
        derived_uncertainties["k_p"] = np.sqrt(
            (k_err * (1 + sym)) ** 2 + (k * sym_err) ** 2
        )
        derived_uncertainties["k_m"] = np.sqrt(
            (k_err * (1 - sym)) ** 2 + (k * sym_err) ** 2
        )

        return derived_params, derived_uncertainties
